//------------------------------------------------------------------------------
//The contents of this file are subject to the Mozilla Public License
//Version 1.1 (the "License"); you may not use this file except in compliance
//with the License. You may obtain a copy of the License at
//http://www.mozilla.org/MPL/ Software distributed under the License is
//distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
//or implied. See the License for the specific language governing rights and
//limitations under the License.
//
//The Original Code is MmeStrm.pas.
//
//The Initial Developer of the Original Code is Alex Shovkoplyas, VE3NEA.
//Portions created by Alex Shovkoplyas are
//Copyright (C) 2007 Alex Shovkoplyas. All Rights Reserved.
//------------------------------------------------------------------------------




unit MmeStrm;

interface

uses
  Windows, Messages, SysUtils, Classes, Controls, SndTypes, MMSystem, Forms,
  SndConv;


type
  TDataDirections = (ddCapture, ddRender, ddBoth);


  TDataEvent = procedure (Sender: TObject; Int32Buffer: TDataBufferI;
    InBuffer, OutBuffer: TDataBufferF) of object;


  PWaveFormatExtensible = ^TWaveFormatExtensible;
  TWaveFormatExtensible = record
    Format: TWaveFormatEx;
    Samples: Word;
    dwChannelMask: DWORD;
    SubFormat: TGUID;
  end;


  PWaveBuffer = ^TWaveBuffer;
  TWaveBuffer = record
    Hdr: TWaveHdr;
    Data: TByteArray;
    end;


  //this class makes settings available at design time
  TDeviceSettings = class(TPersistent)
  private
    FIndex: integer;
    FChannels: integer;
    FSamplingRate: integer;
    FBitsPerSample: integer;
    FSamplesPerBuf: integer;
  protected
    procedure AssignTo(Dest: TPersistent); override;
  public
    constructor Create;
  published
    property Index: integer read FIndex write FIndex;
    property Channels: integer read FChannels write FChannels;
    property SamplingRate: integer read FSamplingRate write FSamplingRate;
    property BitsPerSample: integer read FBitsPerSample write FBitsPerSample;
    property SamplesPerBuf: integer read FSamplesPerBuf write FSamplesPerBuf;
    end;


  TMmeAudioStream = class(TComponent)
  private
    FDirections: TDataDirections;
    FInputDevice: TDeviceSettings;
    FOutputDevice: TDeviceSettings;
    FIsRunning: boolean;
    FBufferCount: integer;
    FOnData: TDataEvent;

    FmtI, FmtO: TWaveFormatExtensible;
    InH, OutH, WinH: THandle;
    rc: MMRESULT;

    InBufs, OutBufs: array of TWaveBuffer;
    RingBuf: array[0..3] of TByteArray;
    RdIdx, WrIdx: integer;
    FloatInBuf, FloatOutBuf: TDataBufferF;

    procedure SndWndProc(var Msg: TMessage);
    procedure WmData(var Msg: TMessage);
    procedure WmDone(var Msg: TMessage);
    procedure CheckErr;
    procedure AddInBuffer(Buf: PWaveBuffer);
    procedure AddOutBuffer(Buf: PWaveBuffer; Data: TByteArray);
    procedure StartInput;
    procedure StartOutput;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure Open;
    procedure Close;
    property IsRunning: boolean read FIsRunning;
  published
    property Directions: TDataDirections read FDirections write FDirections;
    property InputDevice: TDeviceSettings read FInputDevice write FInputDevice;
    property OutputDevice: TDeviceSettings read FOutputDevice write FOutputDevice;
    property BufferCount: integer read FBufferCount write FBufferCount;
    property OnData: TDataEvent read FOnData write FOnData;
  end;

procedure Register;


var
  MmeInDevices: TStringList;
  MmeOutDevices: TStringList;




implementation

procedure Register;
begin
  RegisterComponents('Snd', [TMmeAudioStream]);
end;



//------------------------------------------------------------------------------
//                          device settings object
//------------------------------------------------------------------------------

{ TDeviceSettings }

constructor TDeviceSettings.Create;
begin
  Index := 0;
  Channels := 2;
  SamplingRate := 48000;
  BitsPerSample := 24;
  SamplesPerBuf := 2048;
end;


procedure TDeviceSettings.AssignTo(Dest: TPersistent);
var
  Dst: TDeviceSettings;
begin
  if Dest is TDeviceSettings
    then Dst := TDeviceSettings(Dest)
    else raise Exception.Create('Invalid assignment to DeviceSettings');

  Dst.FChannels := FChannels;
  Dst.FBitsPerSample := FBitsPerSample;
  Dst.FIndex := FIndex;
  Dst.FSamplesPerBuf := FSamplesPerBuf;
  Dst.FSamplingRate := FSamplingRate;
end;





{ TMmeAudioStream }

//------------------------------------------------------------------------------
//                              system
//------------------------------------------------------------------------------
constructor TMmeAudioStream.Create(AOwner: TComponent);
begin
  inherited;

  Directions := ddBoth;

  FInputDevice := TDeviceSettings.Create;
  FOutputDevice := TDeviceSettings.Create;

  WinH := AllocateHWnd(SndWndProc);

  BufferCount := 6;
end;


destructor TMmeAudioStream.Destroy;
begin
  Close;
  DeallocateHWnd(WinH);
  FInputDevice.Free;
  FOutputDevice.Free;

  inherited;
end;


procedure TMmeAudioStream.CheckErr;
var
  Buf: array [0..MAXERRORLENGTH-1] of Char;
begin
  if rc = MMSYSERR_NOERROR then Exit;

  if waveInGetErrorText(rc, Buf, MAXERRORLENGTH) = MMSYSERR_NOERROR
    then raise Exception.Create(Buf)
    else raise Exception.Create('Unknown error: ' + IntToStr(rc));
end;








//------------------------------------------------------------------------------
//                              start/stop
//------------------------------------------------------------------------------
const
  KSDATAFORMAT_SUBTYPE_PCM: TGUID = '{00000001-0000-0010-8000-00aa00389b71}';
  WAVE_FORMAT_EXTENSIBLE = $FFFE;


procedure TMmeAudioStream.Open;
begin
  if FIsRunning then Exit;
  FIsRunning := true;

  InBufs := nil; OutBufs := nil;
  SetLength(InBufs, FBufferCount);
  SetLength(OutBufs, FBufferCount);
  WrIdx := 0;
  RdIdx := Length(RingBuf) div 2;

  FloatInBuf := nil; FloatOutBuf := nil;
  SetLength(FloatInBuf, FInputDevice.FChannels, FInputDevice.FSamplesPerBuf);
  SetLength(FloatOutBuf, FOutputDevice.FChannels, FOutputDevice.FSamplesPerBuf);

  try
    if FDirections in [ddCapture, ddBoth] then StartInput;
    if FDirections in [ddRender, ddBoth] then StartOutput;
  except
    try Close; except end;
    raise;
  end;
end;


procedure TMmeAudioStream.StartInput;
var
  i: integer;
begin
  //fill fmt structure
  with FmtI.Format do
    begin
    wFormatTag        := WAVE_FORMAT_PCM;//WAVE_FORMAT_EXTENSIBLE;
    nChannels         := FInputDevice.FChannels;
    nSamplesPerSec    := FInputDevice.FSamplingRate;
    wBitsPerSample    := FInputDevice.FBitsPerSample;
    nBlockAlign       := (FInputDevice.FBitsPerSample div 8) * nChannels;
    nAvgBytesPerSec   := nSamplesPerSec * nBlockAlign;
    cbSize            := SizeOf(TWaveFormatExtensible) - SizeOf(TWaveFormatEx);
    end;
  FmtI.Samples := FInputDevice.FBitsPerSample;
  FmtI.dwChannelMask := 0;
  FmtI.SubFormat := KSDATAFORMAT_SUBTYPE_PCM;

  //start input
  rc := waveInOpen(@InH, FInputDevice.Index, @FmtI, WinH, 0, CALLBACK_WINDOW);
  CheckErr;
  for i:=0 to High(InBufs) do AddInBuffer(@InBufs[i]);
  rc := waveInStart(InH);
  CheckErr;
end;


procedure TMmeAudioStream.StartOutput;
var
  i: integer;
  EmptyBuf: TByteArray;
begin
  //fill fmt structure
  with FmtO.Format do
    begin
    wFormatTag        := WAVE_FORMAT_EXTENSIBLE;
    nChannels         := FOutputDevice.FChannels;
    nSamplesPerSec    := FOutputDevice.FSamplingRate ;
    wBitsPerSample    := FOutputDevice.FBitsPerSample;
    nBlockAlign       := (FOutputDevice.FBitsPerSample div 8) * nChannels;
    nAvgBytesPerSec   := nSamplesPerSec * nBlockAlign;
    cbSize            := SizeOf(TWaveFormatExtensible) - SizeOf(TWaveFormatEx);
    end;
  FmtO.Samples := FInputDevice.FBitsPerSample;
  FmtO.dwChannelMask := 0;
  FmtO.SubFormat := KSDATAFORMAT_SUBTYPE_PCM;

  //start output
  rc := waveOutOpen(@OutH, FOutputDevice.Index, @FmtO, WinH, 0, CALLBACK_WINDOW);
  CheckErr;
  SetLength(EmptyBuf, FOutputDevice.FSamplesPerBuf * FmtO.Format.nBlockAlign);
  for i:=0 to High(OutBufs) do AddOutBuffer(@OutBufs[i], EmptyBuf);
end;


procedure TMmeAudioStream.Close;
var
  i: integer;
begin
  if not FIsRunning then Exit;
  FIsRunning := false;

  //input
  if FDirections in [ddCapture, ddBoth] then
    begin
    rc := waveInReset(InH);
    CheckErr;

    for i:=0 to High(InBufs) do
      begin
      if (InBufs[i].Hdr.dwFlags and WHDR_PREPARED) <> 0 then
        rc := WaveInUnprepareHeader(InH, @InBufs[i].Hdr, SizeOf(TWaveHdr));
      CheckErr;
      end;

    rc := waveInClose(InH);
    CheckErr;
    end;


  //output
  if FDirections in [ddRender, ddBoth] then
    begin
    rc := waveOutReset(OutH);
    CheckErr;

    for i:=0 to High(OutBufs) do
      begin
      if (OutBufs[i].Hdr.dwFlags and WHDR_PREPARED) <> 0 then
        rc := WaveOutUnprepareHeader(OutH, @OutBufs[i].Hdr, SizeOf(TWaveHdr));
      CheckErr;
      end;

    rc := waveOutClose(OutH);
    CheckErr;
    end;
end;






//------------------------------------------------------------------------------
//                               buffers
//------------------------------------------------------------------------------
procedure TMmeAudioStream.AddInBuffer(Buf: PWaveBuffer);
begin
  //re-create buffer
  Buf.Data := nil;
  SetLength(Buf.Data, FInputDevice.FSamplesPerBuf * FmtI.Format.nBlockAlign);

  //fill header
  FillChar(Buf.Hdr, SizeOf(TWaveHdr), 0);
  with Buf.Hdr do
    begin
    lpData := @Buf.Data[0];
    dwBufferLength := Length(Buf.Data);
    dwUser := DWORD(Buf);
    end;

  //add buffer
  rc := waveInPrepareHeader(InH, @Buf.Hdr, SizeOf(TWaveHdr));
  CheckErr;
  rc := waveInAddBuffer(InH, @Buf.Hdr, SizeOf(TWaveHdr));
  CheckErr;
end;


procedure TMmeAudioStream.AddOutBuffer(Buf: PWaveBuffer; Data: TByteArray);
begin
  Buf.Data := Copy(Data);

  FillChar(Buf.Hdr, SizeOf(TWaveHdr), 0);
  with Buf.Hdr do
    begin
    lpData := @Buf.Data[0];
    dwBufferLength := Length(Buf.Data);
    dwUser := DWORD(Buf);
    end;

  rc := waveOutPrepareHeader(OutH, @Buf.Hdr, SizeOf(TWaveHdr));
  CheckErr;
  rc := waveOutWrite(OutH, @Buf.Hdr, SizeOf(TWaveHdr));
  CheckErr;
end;







//------------------------------------------------------------------------------
//                                events
//------------------------------------------------------------------------------
procedure TMmeAudioStream.SndWndProc(var Msg: TMessage);
begin
  try
    case Msg.Msg of
      MM_WIM_DATA: WmData(Msg);
      MM_WOM_DONE: WmDone(Msg);
      else Msg.Result := DefWindowProc(WinH, Msg.Msg, Msg.wParam, Msg.lParam);
      end;
  except
    Application.HandleException(Self);
  end
end;


procedure TMmeAudioStream.WmData(var Msg: TMessage);
var
  Buf: PWaveBuffer;
begin
  if not IsRunning then Exit;
  if FDirections = ddRender then Exit;
  if THandle(Msg.wParam) <> InH then Exit;

  Buf := PWaveBuffer(PWaveHdr(Msg.lParam).dwUser);

  //read from input buffer
  ConvertToSingle(Buf.Data, FloatInBuf, @FmtI);

  //send to client for processing
  if Assigned(FOnData) then FOnData(Self, nil, FloatInBuf, FloatOutBuf);

  //write to ring buffer
  SetLength(RingBuf[WrIdx], FOutputDevice.FSamplesPerBuf * FmtO.Format.nBlockAlign);
  ConvertFromSingle(FloatOutBuf, RingBuf[WrIdx], @FmtO);
  Inc(WrIdx);
  if WrIdx = Length(RingBuf) then WrIdx := 0;

  //recycle buffer
  rc := WaveInUnprepareHeader(InH, @Buf.Hdr, SizeOf(TWaveHdr));
  CheckErr;
  Buf.Data := nil;
  AddInBuffer(Buf);
end;


procedure TMmeAudioStream.WmDone(var Msg: TMessage);
var
  Buf: PWaveBuffer;
  Data: TByteArray;
begin
  if not IsRunning then Exit;
  if FDirections = ddCapture then Exit;
  if THandle(Msg.wParam) <> OutH then Exit;

  Buf := PWaveBuffer(PWaveHdr(Msg.lParam).dwUser);

  rc := WaveOutUnprepareHeader(OutH, @Buf.Hdr, SizeOf(TWaveHdr));
  CheckErr;

  if FDirections = ddRender
    then //render only, request data from client
      begin
      if Assigned(FOnData) then FOnData(Self, nil, nil, FloatOutBuf);
      SetLength(Data, FOutputDevice.FSamplesPerBuf * FmtO.Format.nBlockAlign);
      ConvertFromSingle(FloatOutBuf, Data, @FmtO);
      AddOutBuffer(Buf, Data);
      end

    else //duplex, get data from ring buffer
      begin
      AddOutBuffer(Buf, RingBuf[RdIdx]);
      Inc(RdIdx);
      if RdIdx = Length(RingBuf) then RdIdx := 0;
      if RdIdx = WrIdx then
        RdIdx := (RdIdx + Length(RingBuf) div 2) mod Length(RingBuf);
      end;
end;








//------------------------------------------------------------------------------
//                              device list
//------------------------------------------------------------------------------
procedure ListDevices;
var
  i: integer;
  WaveInCaps: TWaveInCaps;
  WaveOutCaps: TWaveOutCaps;
begin
  //input devices
  for i:=0 to waveInGetNumDevs-1 do
    begin
    waveInGetDevCaps(i, @WaveInCaps, sizeof(TWaveInCaps));
    MmeInDevices.Add(Format('%.2d  %s', [i+1, PChar(@WaveInCaps.szPname)]));
    end;

  //output devices
  for i:=0 to waveOutGetNumDevs-1 do
    begin
    waveOutGetDevCaps(i, @WaveOutCaps, sizeof(TWaveOutCaps));
    MmeOutDevices.Add(Format('%.2d  %s', [i+1, PChar(@WaveOutCaps.szPname)]));
    end;
end;





initialization
  MmeInDevices := TStringList.Create;
  MmeOutDevices := TStringList.Create;
  ListDevices;


finalization
  MmeInDevices.Free;
  MmeOutDevices.Free;




end.

